#!/usr/bin/env python3

import pandas as pd
import numpy as np
import itertools

import pickle
from tqdm import tqdm
import json
import sys

df = pd.read_csv(sys.argv[1])
result_file = open(sys.argv[2], 'wb')

all_flags = "SFRPAUECN"
const_feature_names = ["sourceTransportPort","destinationTransportPort","protocolIdentifier"]
ip_total = 'accumulate(ipTotalLength)'
inter_time = 'accumulate(_interPacketTimeNanoseconds)'
direction = 'accumulate(flowDirection)'
flags = "accumulate(_tcpFlags)"

min_len = 10
max_len = 500

def read_list(l):
    if isinstance(l, float) and np.isnan(l): return []
    assert((l[0],l[-1]) == ('[', ']'))
    return l[1:-1].split(' ')

def read_numlist(l):
    return ( [float(item)] for item in read_list(l) )

def read_directionlist(l):
    return ( [int(item == 'true')] for item in read_list(l) )

def read_flaglist(l):
    return ( [ int(f in flags) for f in all_flags] for flags in read_list(l) )

def generate_samedirection_IAT(directions, iats):
    # could vectorize this with cumsum, but rather avoid numerical problems
    iat_sum = [None, None]
    result = np.zeros(len(directions))
    for i,direction,iat in zip(range(len(directions)), directions.astype(int), iats):
        if iat_sum[1-direction] is not None:
            iat_sum[1-direction] += iat
        result[i] = 0 if iat_sum[direction] is None else (iat_sum[direction] + iat)
        iat_sum[direction] = 0
    return result


def read_flow(row):
    global wrote_json
    features = []
    const_features = [ row[item] for item in const_feature_names ]

    generators = []
    generators.append( (const_features for _ in itertools.count() ) )
    generators.append( read_numlist(row[ip_total]) )
    generators.append( ([item[0]/1000000] for item in itertools.chain([[0]], read_numlist(row[inter_time])) ))
    generators.append( read_directionlist(row[direction]) )
    generators.append( read_flaglist(row[flags]) )

    return row['Label'], np.array([ list(itertools.chain(*feats)) for _,feats in zip(range(max_len), zip(*generators)) ], dtype=np.float32)

if __name__=="__main__":
    flows = []

    for i, row in tqdm(df.iterrows(), total=df.shape[0]):
        label, flow = read_flow(row)
        # skip flows that have negative IATs
        if len(flow) >= min_len and np.all(flow[:,4] >= 0):
            flows.append((label, np.concatenate((flow, generate_samedirection_IAT(flow[:,5], flow[:,4])[:,None]), axis=1)))

    for f in flows:
        # TODO: maybe set log-iat for first packet to 0
        # pkt length
        f[1][:,3] = np.log(f[1][:,3])
        # iat
        f[1][:,4] = np.log(f[1][:,4] + 0.001)
        # directional iat
        f[1][:,-1] = np.log(f[1][:,-1] + 0.001)

    all_packets = np.concatenate([ f for _,f in flows ], axis=0)
    means = np.mean(all_packets, axis=0)
    stds = np.std(all_packets, axis=0)
    # ignore irrelevant features
    mask = stds > 0.0
    # for ports it's better if we do not normalize
    # otherwise rare ports would cause tiny stdev and give rise to huge outliers
    means[5:-1] = 0.0
    stds[5:-1] = 1.0

    transforms = np.quantile(all_packets[:,[3,4,-1]], np.arange(0,10_000)/10_000, axis=0)
    del all_packets

    pickle.dump({'data': flows, 'means': means, 'stds': stds, 'mask': mask, 'transforms': transforms}, result_file)
