
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.offsetbox import DrawingArea, TextArea, HPacker, VPacker

from matplotlib.transforms import Affine2D

def _approx_box_height(leg, fontsize):
    """
    Return the approximate height and descent of the DrawingArea
    holding the legend handle.
    """
    # The approximate height and descent of text. These values are
    # only used for plotting the legend handle.
    descent = 0.35 * fontsize * (leg.handleheight - 0.7)
    # 0.35 and 0.7 are just heuristic numbers and may need to be improved.
    height = fontsize * leg.handleheight - descent
    # each handle needs to be drawn inside a box of (x, y, w, h) =
    # (0, -descent, width, height).  And their coordinates should
    # be given in the display coordinates.
    return descent, height

def legend_artist(handler, legend, orig_handle,
                    fontsize, handlebox, rotate=False):
    """
    Return the artist that this HandlerBase generates for the given
    original artist/handle.
    Parameters
    ----------
    legend : `~matplotlib.legend.Legend`
        The legend for which these legend artists are being created.
    orig_handle : :class:`matplotlib.artist.Artist` or similar
        The object for which these legend artists are being created.
    fontsize : int
        The fontsize in pixels. The artists being created should
        be scaled according to the given fontsize.
    handlebox : `matplotlib.offsetbox.OffsetBox`
        The box which has been created to hold this legend entry's
        artists. Artists created in the `legend_artist` method must
        be added to this handlebox inside this method.
    """
    xdescent, ydescent, width, height = handler.adjust_drawing_area(
                legend, orig_handle,
                handlebox.xdescent, handlebox.ydescent,
                handlebox.width, handlebox.height,
                fontsize)
    transform = handlebox.get_transform()
    if rotate:
        point = (width - xdescent) / 2
        rotate_transform = Affine2D().rotate_deg_around(point, point, 90)
        transform = rotate_transform + transform
        width, height = height, width
        xdescent, ydescent = ydescent, xdescent
    artists = handler.create_artists(legend, orig_handle,
                                    xdescent, ydescent, width, height,
                                    fontsize, transform)

    # create_artists will return a list of artists.
    for a in artists:
        handlebox.add_artist(a)

    # we only return the first artist
    return artists[0]

def legend_text(handle, s, fig, leg, pos_fun, rotate=False):
    t = TextArea(s, {'rotation': 'vertical' if rotate else 'horizontal'})
    fontsize = t._text.get_fontproperties().get_size_in_points()

    descent, height = _approx_box_height(leg, fontsize)
    if rotate:
        box_width, box_height = height, leg.handlelength * fontsize
        xdescent, ydescent = descent, 0
    else:
        box_width, box_height = leg.handlelength * fontsize, height
        xdescent, ydescent = 0, descent

    box = DrawingArea(width=box_width, height=box_height, xdescent=xdescent, ydescent=ydescent)

    leg_handler = leg.get_legend_handler(leg.get_legend_handler_map(), handle)
    legend_artist(leg_handler, leg, handle, fontsize, box, rotate=rotate)

    if rotate:
        hp = VPacker(0, sep=leg.handletextpad * fontsize, children=(t,box))
    else:
        hp = HPacker(0, sep=leg.handletextpad * fontsize, children=(box,t))

    box.set_figure(fig)
    t.set_figure(fig)
    hp.set_figure(fig)
    all_boxes[fig] = all_boxes.get(fig, [])
    all_boxes[fig].append((hp, pos_fun, rotate))

def draw_hook(fig, renderer, *args, **kwargs):
    old_draw(fig, renderer, *args, **kwargs)
    for hp,pos_fun,rotate in all_boxes.get(fig, []):
        box_x, box_y = pos_fun()
        box_width,box_height,*_ = hp.get_extent(renderer)
        if rotate:
            hp.set_offset((box_x, box_y-box_height/2))
        else:
            hp.set_offset((box_x-box_width/2, box_y))
        hp.draw(renderer)
    
old_draw = Figure.draw
Figure.draw = draw_hook

all_boxes = {}
