#!/usr/bin/env python3

import numpy as np
import tensorflow as tf
import pickle
import os
import random
import sys
import datetime
import argparse
from operator import itemgetter
from tensorflow import keras
from tensorflow.keras import optimizers

FLOAT_TYPE = np.float64
N_PROJECTIONS = 20
N_HIST_BINS = 50
TRUNCATED_LEN = 100

BATCH_SIZE = 128

VAL_PORTION = 0.1

random.seed(0)
np.random.seed(0)

parser = argparse.ArgumentParser()
parser.add_argument('--mode', choices=['ad', 'separate'], required=True, help='Whether to run in anomaly detection (ad) or flow separation (separate) mode.')
parser.add_argument('--data', type=str, required=True, help='File name of the data set in the data/ directory without the .pickle extension.')
parser.add_argument('--id', type=str, default=None, help='Which name to use for storing models and tensorboard logs.')
parser.add_argument('--zero-feat', nargs='+', type=int, help='Features to ignore by zeroing them.')
parser.add_argument('--reverse', action='store_true', help='Reverse time dimension of flows.')
opt = parser.parse_args()

data = pickle.load(open('data/%s.pickle' % opt.data, 'rb'))
feature_mask = data['mask']
input_dim = np.sum(feature_mask)
flows = [ (f[0], ((f[1][:,feature_mask] - data['means'][None,feature_mask]) / data['stds'][None,feature_mask]).astype(FLOAT_TYPE)) for f in data['data'] ]

for i in opt.zero_feat or []:
    for _,f in flows:
        f[:,i] = 0

if opt.reverse:
    assert opt.mode == 'separate' # currently necessary due to feature indices
    def reverse_flow_data(f):
        # need to truncate to TRUNCATED_LEN-2 because in solve.py we do that too
        f = f[TRUNCATED_LEN-3::-1]
        mask = f[:,2] == 1
        # fix IATs
        first_marker = f[-1,1]
        f[1:,1] = f[:-1,1]
        f[0,1] = first_marker

        # fix directional IATs
        dir_iat = f[mask,3]
        if dir_iat.size:
            first_marker = dir_iat[-1]
            dir_iat[1:] = dir_iat[:-1]
            dir_iat[0] = first_marker
            f[mask,3] = dir_iat
        dir_iat = f[~mask,3]
        if dir_iat.size:
            first_marker = dir_iat[-1]
            dir_iat[1:] = dir_iat[:-1]
            dir_iat[0] = first_marker
            f[~mask,3] = dir_iat
        return f
    for i in range(len(flows)):
        flows[i] = (flows[i][0], reverse_flow_data(flows[i][1]) )

perm = np.random.RandomState(0).permutation(len(flows))

if opt.mode == 'ad':
    flows_train = [ flows[i] for i in perm[int(len(flows)*VAL_PORTION):] if flows[i][0] != 1 ]
    flows_val = [ flows[i] for i in perm[:int(len(flows)*VAL_PORTION)] if flows[i][0] != 1 ]
    # dont use port numbers for random projections
    SKIP_FEAT_FOR_PROJ = 2 
elif opt.mode == 'separate':
    flows_train = [ flows[i] for i in perm[int(len(flows)*VAL_PORTION):] ]
    flows_val = [ flows[i] for i in perm[:int(len(flows)*VAL_PORTION)] ]
    SKIP_FEAT_FOR_PROJ = 0

model_id = opt.id or datetime.datetime.now().strftime(r'%Y%m%d-%H%M%S')

def get_proj_minmax():
    proj_maxs = np.zeros(2*N_PROJECTIONS, dtype=FLOAT_TYPE)
    proj_mins = np.zeros(2*N_PROJECTIONS, dtype=FLOAT_TYPE)
    for proj_ind in range(N_PROJECTIONS):
        projected = []
        projected_2pkt = []
        for f in flows:
            projected.append(np.dot(f[1][:,SKIP_FEAT_FOR_PROJ:], projection_matrix1[:,proj_ind]))
            if f[1].shape[0] > 1:
                projected_2pkt.append(np.dot(np.hstack((f[1][:-1,SKIP_FEAT_FOR_PROJ:-1], f[1][1:,SKIP_FEAT_FOR_PROJ:-1])), projection_matrix2[:,proj_ind]))
        proj_mins[proj_ind], proj_maxs[proj_ind] = np.quantile(np.hstack(projected), [0.02, 0.98])
        proj_mins[N_PROJECTIONS+proj_ind], proj_maxs[N_PROJECTIONS+proj_ind] = np.quantile(np.hstack(projected_2pkt), [0.02, 0.98])
    return proj_mins[None,:], proj_maxs[None,:]

projection_matrix1 = np.random.normal(size=(input_dim-SKIP_FEAT_FOR_PROJ, N_PROJECTIONS)).astype(FLOAT_TYPE)
projection_matrix2 = np.random.normal(size=((input_dim-SKIP_FEAT_FOR_PROJ)*2-2, N_PROJECTIONS)).astype(FLOAT_TYPE)
proj_mins, proj_maxs = get_proj_minmax()

def leaky_relu(x):
    return keras.activations.relu(x, alpha=0.3)

def build_model():
    spans = np.max([ np.max(f[1], axis=0) for f in flows ], axis=0) - np.min([ np.min(f[1], axis=0) for f in flows ], axis=0)
    spans[spans == 0] = 1
    def layer1_initializer(shape, dtype=None):
        # simulate min-max scaling of data
        return keras.initializers.GlorotUniform()(shape, dtype) / spans[:,None]
    x = keras.Input(shape=(None, input_dim))
    # masked = keras.layers.Masking(mask_value=0.0)(x)
    hidden = keras.layers.Dense(16, activation=leaky_relu, kernel_initializer=layer1_initializer)(x)
    lstm = keras.layers.LSTM(32, return_sequences=True)(hidden)
    hidden2 = keras.layers.Dense(64, activation=leaky_relu)(lstm)
    lstm2 = keras.layers.LSTM(128, return_sequences=True)(hidden2)
    hidden3 = keras.layers.Dense(256, activation=leaky_relu)(lstm2)
    lstm3 = keras.layers.LSTM(512, return_sequences=True)(hidden3)
    hidden4 = keras.layers.Dense(1024, activation=leaky_relu)(lstm3)
    dense = [ keras.layers.Dense(N_HIST_BINS+1, activation='softmax')(hidden4) for _ in range(2*N_PROJECTIONS) ]
    model = keras.Model(inputs=x, outputs=tf.stack(dense, axis=2))

    def loss(y_true, y_pred):
        # mark ignored time steps with y_true==-1, since tensorflow seems to ignore
        # the mask for loss computation here
        mask = y_true[:,:,0] != -1
        masked_loss = \
            [ tf.cast(mask, tf.float32) * keras.losses.sparse_categorical_crossentropy(tf.maximum(y_true[:,:,i], np.int64(0)), y_pred[:,:,i]) for i in range(2*N_PROJECTIONS) ]
        return  1/((2*N_PROJECTIONS)**.5) * 1/tf.reduce_sum(tf.cast(mask, tf.float32), 1) * tf.reduce_sum(masked_loss, axis=[0,2])

    model.compile(loss=loss, optimizer=optimizers.Adam(learning_rate=0.0001))
    return model

model = build_model()

def gen(flow_data, batch_size):
    mins = (proj_mins[:,:N_PROJECTIONS], proj_mins[:,N_PROJECTIONS:])
    spans = (proj_maxs[:,:N_PROJECTIONS]-mins[0], proj_maxs[:,N_PROJECTIONS:]-mins[1])
    while True:
        perm = np.random.permutation(len(flow_data))
        for i in range(0, len(flow_data), batch_size):
            flow_slice = itemgetter(*perm[i:i+batch_size])(flow_data)
            maxlen = 1 + max(f[1].shape[0] for f in flow_slice)
            batch_X = np.zeros((len(flow_slice), maxlen, input_dim), dtype=FLOAT_TYPE)
            batch_y = np.zeros((len(flow_slice), maxlen, 2*N_PROJECTIONS), dtype=int)
            for i,(_,f) in enumerate(flow_slice):
                batch_X[i,1:f.shape[0]+1] = f
                batch_y[i,:f.shape[0],:N_PROJECTIONS] = np.maximum(1, np.minimum(N_HIST_BINS, ((1+N_HIST_BINS) * (np.dot(f[:,SKIP_FEAT_FOR_PROJ:], projection_matrix1) - mins[0]) / spans[0]).astype(int) ))
                batch_y[i,:f.shape[0]-1,N_PROJECTIONS:] = np.maximum(1, np.minimum(N_HIST_BINS, ((1+N_HIST_BINS) * (np.dot(np.hstack((f[:-1,SKIP_FEAT_FOR_PROJ:-1], f[1:,SKIP_FEAT_FOR_PROJ:-1])), projection_matrix2) - mins[1]) / spans[1]).astype(int) ))
                batch_y[i,(f.shape[0]-1 if opt.reverse else f.shape[0]+1):] = -1
                # forward direction: end in last packet in flow is marked with 0 except the flow's been truncated
            if opt.reverse:
                yield batch_X[:,:TRUNCATED_LEN+1], batch_y[:,:TRUNCATED_LEN+1]
            else:
                yield batch_X[:,:TRUNCATED_LEN-1], batch_y[:,:TRUNCATED_LEN-1]

class storeModelCallback(keras.callbacks.Callback):
    def __init__(self):
        pass

    def on_epoch_end(self, epoch, logs=None):
        if epoch % 5 == 0:
            pickle.dump(((feature_mask, projection_matrix1, projection_matrix2, proj_mins, proj_maxs), model.get_weights()), open('models/%s/%d.pickle' % (model_id, epoch), 'wb'))

os.makedirs('models/%s' % model_id, exist_ok=True)
tb_callback = tf.keras.callbacks.TensorBoard(log_dir='logs/' + model_id)
model.fit(
    gen(flows_train, BATCH_SIZE),
    steps_per_epoch=int(np.ceil(len(flows_train)/BATCH_SIZE)),
    epochs=10000,
    shuffle=False,
    validation_data=gen(flows_val, 512),
    validation_steps=int(np.ceil(len(flows_val)/512)),
    verbose=2,
    callbacks=[storeModelCallback(), tb_callback]
)
