#!/usr/bin/env python3

import pandas as pd
import numpy as np
from tqdm import tqdm
import sys
import itertools
import pickle

min_len = 10
max_len = 500

def read_list(l):
    if isinstance(l, float) and np.isnan(l): return []
    assert (l[0],l[-1]) == ('[', ']')
    return l[1:-1].split(' ')

def read_numlist(l):
    return ( [float(item)] for item in read_list(l) )

def read_directionlist(l):
    return ( [int(item == 'true')] for item in read_list(l) )

def generate_samedirection_IAT(directions, iats):
    # could vectorize this with cumsum, but rather avoid numerical problems
    iat_sum = [None, None]
    result = np.zeros(len(directions))
    for i,direction,iat in zip(range(len(directions)), directions.astype(int), iats):
        if iat_sum[1-direction] is not None:
            iat_sum[1-direction] += iat
        result[i] = 0 if iat_sum[direction] is None else (iat_sum[direction] + iat)
        iat_sum[direction] = 0
    return result

def read_flow(row):
    generators = []
    generators.append( read_numlist(row['accumulate(ipTotalLength)']) )
    generators.append( [item[0]/1000000] for item in itertools.chain([[0]], read_numlist(row['accumulate(_interPacketTimeNanoseconds)'])) )
    generators.append( read_directionlist(row['accumulate(flowDirection)']) )

    constant_features = [ row['sourceTransportPort'], row['destinationTransportPort'] ]
    return constant_features, np.array([ list(itertools.chain(*feats)) for _,feats in zip(range(max_len), zip(*generators)) ])


def open_csv(fn):
    # Skip any attack traffic, since this in unlikely to 
    # be representative for real traffic we're interested in
    df = pd.read_csv(fn)
    if 'Label' in df:
        df = df[df['Label'] == 0]
    return df


if __name__=="__main__":
    assert len(sys.argv) >= 3
    result_file = open(sys.argv[2], 'wb')
    flows = []
    for i, row in itertools.chain(*[ tqdm(open_csv(fn).iterrows()) for fn in sys.argv[1:-1] ]):
        constant_features, pkt_features = read_flow(row)
        if len(pkt_features) >= min_len and np.all(pkt_features[:,1] >= 0):
            flows.append((constant_features, np.concatenate((pkt_features, generate_samedirection_IAT(pkt_features[:,2], pkt_features[:,1])[:,None]), axis=1)))

    for f in flows:
        # TODO: maybe set log-iat for first packet to 0
        # pkt length
        f[1][:,0] = np.log(f[1][:,0])
        # iat
        f[1][:,1] = np.log(f[1][:,1] + 0.001)
        # directional iat
        f[1][:,3] = np.log(f[1][:,3] + 0.001)

    all_packets = np.concatenate([ f for _,f in flows ], axis=0)
    means = np.mean(all_packets, axis=0)
    stds = np.std(all_packets, axis=0)
    # ignore irrelevant features
    mask = stds > 0.0
    del all_packets

    # mins = np.min([ np.min(f, axis=0) for _,f in flows ], axis=0)
    # maxs = np.max([ np.max(f, axis=0) for _,f in flows ], axis=0)

    pickle.dump({'data': flows, 'means': means, 'stds': stds, 'mask': mask}, result_file)
