#!/usr/bin/env python3

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib import gridspec
import json
import pandas as pd
import os
from sklearn.metrics import adjusted_rand_score
import numpy as np
import argparse
import pickle

from matplotlib.patches import Rectangle
from matplotlib.legend import Legend
from axlegend import legend_text


VAL_PORTION = 0.1

runs = pd.DataFrame.from_dict(json.load(open('runs.json')), 'index')
for feature in 'wiat', 'reverse', 'skip_2pkt':
    if not feature in runs:
        runs[feature] = False
    runs[feature] = runs[feature] == True

if 'only_steady' not in runs:
    runs['only_steady'] = None

parser = argparse.ArgumentParser()
parser.add_argument('--data', type=str, help='File name of the data set in the data/ directory without the .pickle extension.')
parser.add_argument('--runs', type=str, action='extend', nargs='+')
parser.add_argument('--function', choices=['maxpos_plot', 'acc_plot', 'acc_table', 'steady_plots', 'hist', 'hist2', 'latex_table'], required=True, help='The function that will be called')
parser.add_argument('--wiat', action='store_true')
opt = parser.parse_args()

os.makedirs('plots', exist_ok=True)
default_colors = matplotlib.rcParams['axes.prop_cycle'].by_key()['color']

def offset_ax_coord(ax_pos, offset):
    x, y = plt.gca().transAxes.transform(ax_pos)
    offset2 = plt.gcf().dpi_scale_trans.transform(offset)
    return x+offset2[0], y+offset2[1]

def index_of_next_equal(x):
    res = np.zeros(len(x), dtype=int)
    ar = np.arange(len(x))
    for i in np.unique(x):
        mask = x == i
        res[mask] = np.hstack((ar[mask][1:], -1))
    return res

def iterate_flow_metrics(run, filter_fn=None):
    for line in open(f'results-{opt.data}/run{run}.jsonl', 'r'):
        line_data = json.loads(line)
        if filter_fn is not None and not filter_fn(*line_data):
            continue
        flow_indices, flow_offset_times, duration, loss, max_position, best_config, labels = line_data
        best_config = np.array(best_config)
        labels = np.array(labels)
        acc = np.mean(labels == best_config)
        tr_acc = np.mean(index_of_next_equal(best_config) == index_of_next_equal(labels))
        yield len(labels), acc, tr_acc, adjusted_rand_score(labels, best_config)

def get_metrics(run, flow_averaging=True, filter_fn=None):
    total_packets = 0
    total_flows = 0
    metrics = np.zeros(3)

    for pkt_cnt, *seq_metrics in iterate_flow_metrics(run, filter_fn):
        total_packets += pkt_cnt
        total_flows += 1
        if flow_averaging:
            metrics += seq_metrics
        else:
            metrics += pkt_cnt * np.asarray(seq_metrics)
    metrics /= total_flows if flow_averaging else total_packets
    return *metrics,


# def get_metrics(run, flow_averaging=True, filter_fn=None):
#     total_packets = 0

#     total_acc_sum = 0.0
#     total_tracc_sum = 0.0

#     for line in open(f'results/run{run}.jsonl', 'r'):
#         line_data = json.loads(line)
#         if filter_fn is not None and not filter_fn(*line_data):
#             continue
#         flow_indices, flow_offset_times, duration, loss, max_position, best_config, labels = line_data
#         total_packets += len(labels)
#         best_config = np.array(best_config)
#         labels = np.array(labels)
#         acc = np.mean(labels == best_config)
#         tr_acc = np.mean(index_of_next_equal(best_config) == index_of_next_equal(labels))

#         if flow_averaging:
#             total_acc_sum += acc
#             total_tracc_sum += tr_acc
#         else:
#             total_acc_sum += len(labels) * acc
#             total_tracc_sum += len(labels) * tr_acc

#     if flow_averaging:
#         total_acc_sum /= len(results)
#         total_tracc_sum /= len(results)
#     else:
#         total_acc_sum /= total_packets
#         total_tracc_sum /= total_packets
#     return total_acc_sum, total_tracc_sum

def hist():
    for run, row in runs.iterrows():
        try:
            metrics = np.array(list(iterate_flow_metrics(run)))
        except (FileNotFoundError, ZeroDivisionError):
            print('No results for run %s found.' % run)
            continue
        plt.hist(metrics[:,1])
        plt.xlabel('Accuracy')
        plt.ylabel('Frequency')
        plt.savefig('plots/hist_acc_%s.pdf' % run)
        plt.close()
        plt.hist(metrics[:,2])
        plt.xlabel('TrAccuracy')
        plt.ylabel('Frequency')
        plt.savefig('plots/hist_tracc_%s.pdf' % run)
        plt.close()

def hist2():
    for (label, fn,metric_index), color in zip([('Accuracy', 'hist2_acc.pdf', 1), ('TrAccuracy', 'hist2_tracc.pdf', 2)], default_colors):
        fig, axs = plt.subplots(nrows=len(opt.runs), figsize=(3,3), sharex=True)
        for ax,run in zip(axs, opt.runs):
            metrics = np.array(list(iterate_flow_metrics(run)))
            ax.hist(metrics[:,metric_index], weights=1/len(metrics)*np.ones(len(metrics)), color=color)
        
        # fig.add_subplot(111, frameon=False)
        # hide tick and tick label of the big axis
        # plt.tick_params(labelcolor='none', which='both', top=False, bottom=False, left=False, right=False)
        plt.xlabel(label)
        fig.supylabel('Relative frequency')
        plt.tight_layout()
        plt.subplots_adjust(hspace=.0)
        plt.savefig('plots/%s'%fn, bbox_inches='tight', pad_inches=0)
        plt.close()

def maxpos_plot():
    all_x = []
    default_colors.reverse()
    quantiles = [0.95, 0.9, 0.75, 0.5]
    def do_plot(df, **kwargs):

        x, all_y = [], []
        for run,row in df.iterrows():
            max_positions = []
            results = [ json.loads(line) for line in open(f'results-{opt.data}/run{run}.jsonl', 'r') ]
            for flow_indices, flow_offset_times, duration, loss, max_position, best_config, labels in results:
                if max_position > -1:
                    max_positions.append(max_position+1)
            x.append(int(row['interleave-cnt']))
            all_y.append(np.quantile(max_positions, quantiles).astype(int))
        for color, y in zip(default_colors, np.stack(all_y, axis=1)):
            plt.semilogy(x, y, c=color, **kwargs)
        all_x.extend(x)

    plt.figure(figsize=(5.5,2.9))
    filtered_runs = runs[~(runs['wiat'] | runs['reverse'] | runs['skip_2pkt'] | runs['only_steady'])]
    filtered_runs = filtered_runs.groupby('interleave-cnt').tail(1).sort_values('interleave-cnt')
    print('Selected runs w/ IAT:\n', filtered_runs)
    do_plot(filtered_runs)

    filtered_runs = runs[runs['wiat'] & ~(runs['reverse'] | runs['skip_2pkt'] | runs['only_steady'])]
    filtered_runs = filtered_runs.groupby('interleave-cnt').tail(1).sort_values('interleave-cnt')
    print('Selected runs w/o IAT:\n', filtered_runs)
    do_plot(filtered_runs, linestyle='dashed')

    filtered_runs = runs[runs['skip_2pkt'] & ~(runs['wiat'] | runs['reverse'] | runs['only_steady'])]
    filtered_runs = filtered_runs.groupby('interleave-cnt').tail(1).sort_values('interleave-cnt')
    print('Selected runs w/o 2-pkt projections:\n', filtered_runs)
    do_plot(filtered_runs, linestyle='dashdot')

    plt.xlabel('Interleaved flows')
    plt.ylabel('Maximum rank $\\bar{\\rho}$')
    plt.xticks(np.unique(all_x))

    # add a dummy entry to ensure all percentiles are in one column
    lines = [ Line2D([0], [0], **prop) for prop in
             [*[ {'c': c} for c in default_colors[:len(quantiles)]], {'linestyle': 'dashdot', 'c': 'k'}, {'linestyle': 'dashed', 'c': 'k'}, {'alpha': 0}] ]

    labels = ['$\\mathregular{%d^{th}}$ percentile' % (int(q*100)) for q in quantiles] + ['No 2-packet proj.', 'IAT omitted', ' ' ]
    labels[quantiles.index(0.75)] = '$\\mathregular{3^{rd}}$ quartile'
    labels[quantiles.index(0.5)] = 'Median'
    plt.legend(lines, labels, ncol=2, frameon=False, borderaxespad=0.2)
    plt.ylim((plt.ylim()[0], 40_000))
    plt.tight_layout()
    plt.savefig('plots/maxpos.pdf')

def acc_plot():
    all_x = []
    def do_plot(df, **kwargs):
        x, all_y = [], []
        for run,row in df.iterrows():
            try:
                all_y.append(get_metrics(run))
                x.append(int(row['interleave-cnt']))
            except (FileNotFoundError, ZeroDivisionError):
                pass
        for color, y in zip(default_colors, np.stack(all_y, axis=1)):
            plt.plot(x, y, c=color, **kwargs)
        all_x.extend(x)
    fig = plt.figure(figsize=(5.5,2.5))
    filtered_runs = runs[~(runs['wiat'] | runs['reverse'] | runs['skip_2pkt'] | runs['only_steady'])]
    filtered_runs = filtered_runs.groupby('interleave-cnt').tail(1).sort_values('interleave-cnt')
    print('Selected runs w/ IAT:\n', filtered_runs)
    do_plot(filtered_runs)

    filtered_runs = runs[runs['wiat'] & ~(runs['reverse'] | runs['skip_2pkt'] | runs['only_steady'])]
    filtered_runs = filtered_runs.groupby('interleave-cnt').tail(1).sort_values('interleave-cnt')
    print('Selected runs w/o IAT:\n', filtered_runs)
    do_plot(filtered_runs, linestyle='dashed')

    filtered_runs = runs[runs['skip_2pkt'] & ~(runs['wiat'] | runs['reverse'] | runs['only_steady'])]
    filtered_runs = filtered_runs.groupby('interleave-cnt').tail(1).sort_values('interleave-cnt')
    print('Selected runs w/o 2-pkt projections:\n', filtered_runs)
    do_plot(filtered_runs, linestyle='dashdot')

    filtered_runs = runs[runs['reverse'] & ~(runs['wiat'] | runs['skip_2pkt'] | runs['only_steady'])]
    filtered_runs = filtered_runs.groupby('interleave-cnt').tail(1).sort_values('interleave-cnt')
    print('Selected runs w/ reverse model:\n', filtered_runs)
    do_plot(filtered_runs, linestyle='dotted')

    lines = [ Line2D([0], [0], **prop) for prop in
             [{'linestyle': 'dashed', 'c': 'k'}, {'linestyle': 'dashdot', 'c': 'k'}, {'linestyle': 'dotted', 'c': 'k'},] ]
    plt.legend(lines, ['IAT omitted', 'No 2-packet proj.', 'With reverse model'], frameon=False, borderaxespad=0.2)
    plt.xlabel('Interleaved flows')
    leg = Legend(fig, [], [], handlelength=0.7, handletextpad=0.4)
    legend_text(Rectangle((0,0), 1, 1, fc=default_colors[1], alpha=.8), 'TrAccuracy', fig, leg, lambda: offset_ax_coord((0,0.5), (-.8,.2)), True)
    legend_text(Rectangle((0,0), 1, 1, fc=default_colors[0], alpha=.8), 'Accuracy', fig, leg, lambda: offset_ax_coord((0,0.32), (-.6,.2)), True)
    legend_text(Rectangle((0,0), 1, 1, fc=default_colors[2], alpha=.8), 'ARI', fig, leg, lambda: offset_ax_coord((0,0.8), (-.6,.2)), True)
    # plt.ylabel('Accuracy/TrAccuracy')
    plt.xticks(np.unique(all_x))
    # plt.yticks([.85, 0.9, 0.95, 1.0])
    plt.ylim((0.8, 1.005))
    plt.tight_layout()
    plt.subplots_adjust(left=0.18)
    plt.savefig('plots/accuracy.pdf', pad_inches=0)

def steady_plots():
    flows = pickle.load(open('data/%s.pickle' % opt.data, 'rb'))['data']
    perm = np.random.RandomState(0).permutation(len(flows))
    flows_val = [ flows[i] for i in perm[:int(len(flows)*VAL_PORTION)] ]

    def do_plot(run, **kwargs):
        def my_filter(flow_indices, flow_offset_times, duration, loss, max_position, best_config, labels):
            times1 = np.cumsum(np.exp(flows_val[flow_indices[0]][1][:,1]) - 0.001)
            # print(times1 % 50)
            times2 = np.cumsum(np.exp(flows_val[flow_indices[1]][1][:,1]) - 0.001) + flow_offset_times[0]
            return 0 <= (np.mean(times1) - np.mean(times2) - mod_offset_min) % 50 < (mod_offset_max-mod_offset_min) % 50

        mod_offset_bins = 10
        x, all_y = [], []
        for i in range(-25,25):
            x.append(i*50/50)
            mod_offset_min = x[-1] - 50/mod_offset_bins/2
            mod_offset_max = x[-1] + 50/mod_offset_bins/2
            all_y.append(get_metrics(run, filter_fn=my_filter))
        for color, y in zip(default_colors, np.stack(all_y, axis=1)):
            plt.plot(x, y, c=color, **kwargs)
    for interleave_cnt in np.unique(runs['interleave-cnt']):
        filtered_runs = runs[(runs['interleave-cnt'] == interleave_cnt) & runs['only_steady']]
        if filtered_runs.empty:
            print('No results for interleave-cnt==%d' % interleave_cnt)
        else:
            print('Selected runs:\n', filtered_runs)
            fig = plt.figure(figsize=(5.5,1.9))
            for run in filtered_runs[filtered_runs['only_steady'] == 'samesizes'].tail(1).index:
                do_plot(run)
            for run in filtered_runs[filtered_runs['only_steady'] == 'diffsizes'].tail(1).index:
                do_plot(run, linestyle='dashed')

            lines = [ Line2D([0], [0], **prop) for prop in
                    [{'linestyle': 'solid', 'c': 'k'}, {'linestyle': 'dashed', 'c': 'k'},] ]
            plt.legend(lines, ['Equal pkt. sizes', 'Distinct pkt. sizes'], frameon=False, borderaxespad=0.2)
            plt.xlabel('Time difference (ms)')
            # plt.ylabel('Accuracy/TrAccuracy')
            leg = Legend(fig, [], [], handlelength=0.7, handletextpad=0.4)
            legend_text(Rectangle((0,0), 1, 1, fc=default_colors[1], alpha=.8), 'TrAccuracy', fig, leg, lambda: offset_ax_coord((0,0.5), (-.75,.2)), True)
            legend_text(Rectangle((0,0), 1, 1, fc=default_colors[0], alpha=.8), 'Accuracy', fig, leg, lambda: offset_ax_coord((0,0.28), (-.55,.2)), True)
            legend_text(Rectangle((0,0), 1, 1, fc=default_colors[2], alpha=.8), 'ARI', fig, leg, lambda: offset_ax_coord((0,0.85), (-.55,.2)), True)
            # plt.ylim((plt.ylim()[0], 1.005))
            plt.tight_layout()
            plt.subplots_adjust(left=0.18)
            plt.savefig('plots/steady_%d.pdf' %interleave_cnt)
            plt.close()

def acc_table():
    for run, row in runs.iterrows():
        try:
            print(get_metrics(run))
        except (FileNotFoundError, ZeroDivisionError):
            print('No results for run %s found.' % run)


def latex_table():
    print(' & \\multicolumn{3}{c}{\\textbf{Pkt. averaging}} & \\multicolumn{3}{c}{\\textbf{Seq. averaging}} \\\\')
    print ('\\textbf{Flows} & \\textbf{Acc.} & \\textbf{TrAcc.} & \\textbf{ARI} & \\textbf{Acc.} & \\textbf{TrAcc.} & \\textbf{ARI} \\\\ \\hline')
    for run in opt.runs:
        try:
            metrics_pkt = get_metrics(run, flow_averaging=False)
            metrics_flow = get_metrics(run)
            print('&', runs.loc[run]['interleave-cnt'], '&', ' & '.join('%.3f' %x for x in [*metrics_pkt, *metrics_flow]), '\\\\')
        except (FileNotFoundError, ZeroDivisionError):
            print('No results for run %s found.' % run)

globals()[opt.function]()