#!/usr/bin/env python3

import pickle
import sys
import numpy as np
import random
import json
import os
import fcntl
import gc
import tqdm
import itertools
import argparse
import time
import tensorflow as tf
from tensorflow import keras

N_PROJECTIONS = 20
N_HIST_BINS = 50
APRIORI_DELAY = 5
TRUNCATED_LEN = 100

VAL_PORTION = 0.1

BATCH_SIZE=512

parser = argparse.ArgumentParser()
parser.add_argument('--data', type=str, required=True, help='File name of the data set in the data/ directory without the .pickle extension.')
parser.add_argument('--model', type=str, required=True, help='Path to the model in .pickle format.')
parser.add_argument('--rev-model', type=str, help='Optional path to a model trained with --reverse.')
parser.add_argument('--interleave-cnt', type=int, default=2, help='On how many interleaved flows to operate.')
parser.add_argument('--result-file', type=int, help='Numeric id for file in .jsonl format to write results to.')
parser.add_argument('--retain-configs', type=int, default=2000, help='How many configurations to keep after each time step.')
parser.add_argument('--zero-feat', nargs='+', type=int, help='Features to ignore by zeroing them.')
parser.add_argument('--skip-2pkt', action='store_true', help='Do not evaluate 2-packet projections.')
parser.add_argument('--only-steady', choices=['samesizes', 'diffsizes'])
parser.add_argument('-v', '--verbose', action='store_true')
opt = parser.parse_args()

flows = pickle.load(open('data/%s.pickle' % opt.data, 'rb'))
feature_means = flows['means']
feature_stds = flows['stds']
flows = flows['data']

perm = np.random.RandomState(0).permutation(len(flows))
# need to truncate to TRUNCATED_LEN-2 because in learn.py we train with max. TRUNCATED_LEN-2 steps
# val
flows_val = [ (flows[i][0], flows[i][1][:TRUNCATED_LEN-2,:]) for i in perm[:int(len(flows)*VAL_PORTION)] ]
# train
# flows_val = [ (flows[i][0], flows[i][1][:TRUNCATED_LEN-2,:]) for i in perm[int(len(flows)*VAL_PORTION):] ]

def steady_flow_sampling():
    filtered_indices = [ i for i,f in enumerate(flows_val) if f[1][:,0].min()==f[1][:,0].max() and f[0][1] == 10_001 ]
    pkt_sizes = np.unique([ flows_val[i][1][0,0] for i in filtered_indices ]).tolist()
    if opt.only_steady == 'samesizes':
        selected_size = random.choice(pkt_sizes)
        filtered_indices = [ i for i in filtered_indices if flows_val[i][1][0,0] == selected_size ]
        return [ random.choice(filtered_indices) for _ in range(opt.interleave_cnt) ]
    elif opt.only_steady == 'diffsizes':
        selected_sizes = random.sample(pkt_sizes, opt.interleave_cnt)
        return [ random.choice([ i for i in filtered_indices if flows_val[i][1][0,0] == pkt_size ]) for pkt_size in selected_sizes ]


def get_interleaved_flows():
    if opt.only_steady:
        flow_indices = steady_flow_sampling()
    else:
        flow_indices = [ random.randrange(len(flows_val)) for _ in range(opt.interleave_cnt) ]
    flow_constant, flow_data = zip(*[ flows_val[i] for i in flow_indices ])


    flow_data = [ f.copy() for f in flow_data ]

    flow_data[0][:,1] = np.cumsum(np.exp(flow_data[0][:,1]) - 0.001)
    all_times = flow_data[0][:,1]
    offset_times = []
    for i in range(1,opt.interleave_cnt):
        offset_pkt = random.randrange(5, len(all_times))
        offset_times.append(random.uniform(all_times[offset_pkt-1], all_times[offset_pkt]))
        flow_data[i][:,1] = np.cumsum(np.exp(flow_data[i][:,1]) - 0.001) + offset_times[-1]
        all_times = np.sort(np.hstack((all_times, flow_data[i][:,1])))

    print(flow_constant, offset_times)
    data = np.concatenate(flow_data, axis=0)
    labels = np.concatenate([ i * np.ones(flow_data[i].shape[0]) for i in range(opt.interleave_cnt) ])

    # TODO: maybe avoid having to undo log normalization

    perm = np.argsort(data[:,1], kind='stable')
    data = data[perm,:]
    labels = labels[perm]
    # don't have this feature in practice
    data = np.delete(data, 3, axis=1)
    return flow_indices, offset_times, data, labels.astype(int)


def leaky_relu(x):
    return keras.activations.relu(x, alpha=0.3)


def build_model():
    x = keras.Input(shape=(4,))
    states_in = keras.Input(shape=(2,32+128+512))
    # no masking here
    hidden = keras.layers.Dense(16, activation=leaky_relu)(x)
    lstm, lstm_state = keras.layers.LSTMCell(32)(hidden, (states_in[:,0,:32], states_in[:,1,:32]))
    hidden2 = keras.layers.Dense(64, activation=leaky_relu)(lstm)
    lstm2, lstm2_state = keras.layers.LSTMCell(128)(hidden2, (states_in[:,0,32:32+128], states_in[:,1,32:32+128]))
    hidden3 = keras.layers.Dense(256, activation=leaky_relu)(lstm2)
    lstm3, lstm3_state = keras.layers.LSTMCell(512)(hidden3, (states_in[:,0,32+128:32+128+512], states_in[:,1,32+128:32+128+512]))
    hidden4 = keras.layers.Dense(1024, activation=leaky_relu)(lstm3)
    n_histograms = N_PROJECTIONS if opt.skip_2pkt else 2*N_PROJECTIONS
    dense = [ keras.layers.Dense(N_HIST_BINS+1, activation='softmax')(hidden4) for _ in range(n_histograms) ]

    states_out = tf.concat([tf.stack(lstm_state, 1), tf.stack(lstm2_state, 1), tf.stack(lstm3_state, 1)], 2)
    model = keras.Model(inputs=(x, states_in), outputs=(tf.stack(dense, axis=1), states_out))
    return model


def evaluate_step(previous_tot_loss, previous_associations, packet, flow_states, pkt_association, active_flows, pkt_apriori, new_configurations):
    log_pkt_size, t, pkt_direction = packet
    pkt_direction = int(pkt_direction)
    lstm_state, histograms1, histograms2, last_features, last_t, last_dir_t, debug_info = flow_states[pkt_association]

    # debug_info = debug_info.copy()

    iat = (t - last_t) if last_t is not None else 0
    iat_dir = (t - last_dir_t[pkt_direction]) if last_dir_t[pkt_direction] is not None else 0
    features = np.array((log_pkt_size, np.log(iat + 0.001), pkt_direction, np.log(iat_dir + 0.001)))
    # if pkt_association== 1:
    #     globals()['debug_totflow'].append(features)
    features = (features - feature_means) / feature_stds
    for i in opt.zero_feat or []:
        features[i] = 0

    # if histograms1 is not None: # TODO: no longer possible
    projected = np.dot(features, projection_matrix1)
    bins = ((1+N_HIST_BINS) * (projected - proj_mins[0]) / proj_spans[0]).astype(int)
    bins = np.maximum(np.minimum(bins, N_HIST_BINS), 1)
    loss = -np.sum(np.log(np.maximum(1e-7, histograms1[range(N_PROJECTIONS), bins])))

    if histograms2 is not None and not opt.skip_2pkt:
        projected = np.dot(np.hstack((last_features[:3], features[:3])), projection_matrix2)
        bins = ((1+N_HIST_BINS) * (projected - proj_mins[1]) / proj_spans[1]).astype(int)
        bins = np.maximum(np.minimum(bins, N_HIST_BINS), 1)
        loss += -np.sum(np.log(np.maximum(1e-7, histograms2[range(N_PROJECTIONS,2*N_PROJECTIONS), bins])))
    else:
        loss *= 2
    del projected, bins
    # loss += -np.log(max(1e-20, pkt_apriori))
    loss += -np.log(pkt_apriori)
    lstm_prediction, new_lstm_state = yield features, lstm_state

    # debug_info.append((features, loss, lstm_prediction, new_lstm_state))
    new_last_dir_t = (last_dir_t[0], t) if pkt_direction == 1 else (t, last_dir_t[1])
    new_state = [new_lstm_state, lstm_prediction, histograms1, features, t, new_last_dir_t, debug_info]
    new_configurations.append((
        previous_tot_loss + loss,
        np.hstack((previous_associations, np.uint8(pkt_association))),
        tuple(new_state if i==pkt_association else flow_states[i] for i in range(len(flow_states))),
        max(active_flows, pkt_association+1) ))
    yield

def load_model(fn):
    saved_model = pickle.load(open(fn, 'rb'))
    model = build_model()
    if opt.skip_2pkt:
        # *2 because we have weights and offsets for each histogram
        del saved_model[1][-2*N_PROJECTIONS:]
    model.set_weights(saved_model[1])

    _,projection_matrix1, projection_matrix2, mins, maxs = saved_model[0]
    spans = maxs - mins
    proj_mins = (mins[0,:N_PROJECTIONS], mins[0,N_PROJECTIONS:])
    proj_spans = (spans[0,:N_PROJECTIONS], spans[0,N_PROJECTIONS:])

    initial_histograms, initial_lstm_state = model((np.zeros((1,4)), np.zeros((1,2,32+128+512))))
    return model, projection_matrix1, projection_matrix2, proj_mins, proj_spans, initial_lstm_state.numpy()[0], initial_histograms.numpy()[0]


def update_new_apriori(new_apriori, step_ind, configurations, step_cnt):
    if step_ind < APRIORI_DELAY:
        pkt_indices = []
    elif step_ind == step_cnt-1:
        pkt_indices = list(range(max(0, step_ind-APRIORI_DELAY), step_cnt))
    else:
        pkt_indices = [ step_ind-APRIORI_DELAY ]
    for pkt_ind in pkt_indices:
        min_loss = min(cfg[0] for cfg in configurations)
        for cfg in configurations:
            new_apriori[pkt_ind,cfg[1][pkt_ind]] += np.exp(-cfg[0] + min_loss)
        new_apriori[pkt_ind] /= np.sum(new_apriori[pkt_ind])
        # print(pkt_ind, new_apriori[pkt_ind])
        
def main():
    global projection_matrix1, projection_matrix2
    global proj_mins, proj_spans
    fwd_model = load_model(opt.model)
    rev_model = load_model(opt.rev_model) if opt.rev_model else None

    total_rounds = 0
    total_round_acc_sum = 0.0
    total_round_tracc_sum = 0.0

    total_packets = 0
    total_packet_acc_sum = 0
    total_packet_tracc_sum = 0
    total_loss = 0
    tot_flow_loss = 0

    os.makedirs('debug_info', exist_ok=True)

    result_fn = None if opt.result_file is None else 'results-%s/run%d.jsonl' % (opt.data, opt.result_file)

    if result_fn:
        # create file if needed
        os.makedirs('results-%s' % opt.data, exist_ok=True)
        open(result_fn, 'a')

    for round_counter in itertools.count():
        flow_indices, flow_offset_times, interleaved_data, interleaved_labels = get_interleaved_flows()

        start_time = time.time()
        new_apriori = 1/opt.interleave_cnt * np.ones((len(interleaved_data), opt.interleave_cnt))

        for iteration_counter in range(3 if rev_model else 1):
            is_rev_iteration = iteration_counter%2 == 1
            if is_rev_iteration:
                model, projection_matrix1, projection_matrix2, proj_mins, proj_spans, initial_lstm_state, initial_histograms = rev_model
                data = interleaved_data[::-1].copy()
                data[:,1] *= -1
                labels = interleaved_labels[::-1]
            else:
                model, projection_matrix1, projection_matrix2, proj_mins, proj_spans, initial_lstm_state, initial_histograms = fwd_model
                data = interleaved_data
                labels = interleaved_labels

            max_position = 0
            initial_state = (initial_lstm_state, initial_histograms, None, None, None, (None, None), [])
            configurations = [ (0, np.zeros((0,), dtype=np.uint8), (initial_state,)*opt.interleave_cnt, opt.interleave_cnt if is_rev_iteration else 0) ]

            apriori = new_apriori[::-1]
            new_apriori = np.zeros_like(apriori)

            # globals()['debug_totflow'] = []

            for pkt_ind, (packet, pkt_apriori) in enumerate(zip(tqdm.tqdm(data) if opt.verbose else data, apriori)):
                gc.collect()
                new_configurations = []
                steps = []

                # print(pkt_apriori)

                for tot_loss, associations, states, active_flows in configurations:
                    for i in range(min(opt.interleave_cnt, active_flows+1)):
                        steps.append( evaluate_step(tot_loss, associations, packet, states, i, active_flows, pkt_apriori[i], new_configurations) )

                # Collect model evaluations of all steps into a batch for efficient evaluation
                for i in range(0, len(steps), BATCH_SIZE):
                    batch_steps = steps[i:i+BATCH_SIZE]
                    features_batch, lstm_state_batch = ( tf.stack(x, axis=0) for x in zip(*[ next(step) for step in batch_steps ]) )
                    # print(lstm_state_batch.device)
                    for step, lstm_prediction, new_lstm_state in zip(batch_steps, *model((features_batch, lstm_state_batch))):
                        step.send((lstm_prediction, new_lstm_state))
                    del features_batch, lstm_state_batch

                update_new_apriori(new_apriori, pkt_ind, new_configurations, len(data))

                # configurations = [ x for x in new_configurations if all(a==b for a,b in zip(x[1], labels)) ]
                configurations = sorted(new_configurations, key=lambda x: x[0])[:opt.retain_configs]

                if max_position > -1:
                    for config_ind,config in enumerate(configurations):
                        if all(a==b for a,b in zip(config[1], labels)):
                            max_position = max(max_position, config_ind)
                            break
                    else:
                        max_position = -1
                for config in configurations:
                    for state in config[2]:
                        if isinstance(state[1], tf.Tensor):
                            state[1] = state[1].numpy()

        # print(np.stack(debug_totflow))

        # pickle.dump((flow1_ind, flow2_ind, configurations[0][2][0][-1], configurations[0][2][1][-1]), open('debug_info/%d.pickle' % round_counter, 'wb'))

        done_time = time.time()
        best_config = np.array(configurations[0][1])
        best_perm = max(itertools.permutations(range(opt.interleave_cnt)), key=lambda perm: np.mean(np.array(perm)[best_config] == labels))
        best_config = np.array(best_perm)[best_config]

        if result_fn:
            # use file locking, so that we can safely run on multiple tasks/machines simultaneously
            with open(result_fn, 'r+') as f:
                fcntl.flock(f, fcntl.LOCK_EX)
                f.seek(0,2)
                f.write(json.dumps((flow_indices, flow_offset_times, done_time-start_time, configurations[0][0], max_position, best_config.tolist(), labels.tolist())) + '\n')
        
        acc = np.mean(labels == best_config)
        tr_acc = np.mean( (labels[1:]==labels[:-1]) == (best_config[1:]==best_config[:-1]) )
        total_rounds += 1
        total_round_acc_sum += acc
        total_round_tracc_sum += tr_acc
        total_packets += data.shape[0]
        total_packet_acc_sum += data.shape[0] * acc 
        total_packet_tracc_sum += (data.shape[0]-1) * tr_acc
        total_loss += configurations[0][0]
        tot_flow_loss += configurations[0][0]/data.shape[0]
        if opt.verbose:
            print(labels, best_config, configurations[0][0], max_position)
            print('%.5f %.5f %.5f' % (acc, total_round_acc_sum/total_rounds, total_packet_acc_sum/total_packets))
            print('%.5f %.5f %.5f' % (tr_acc, total_round_tracc_sum/total_rounds, total_packet_tracc_sum/(total_packets-total_rounds)))
            print(total_loss/total_packets, tot_flow_loss/total_rounds)

main()
